/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include "util.h"
#include "getopt.h"
#include "globalregistry.h"
#include "configfile.h"
#include "kis_netclient.h"

// Some system defaults
#define SYSCONF_LOC "/usr/local/etc/"
const char *config_base = "kismet_emailclient.conf";

// Global registry of stuff
GlobalRegistry *globalreg = NULL;

// Some local globals.  Ugh.
char *to_addr = NULL;
char *from_addr = NULL;
char *mail_prog = NULL;
map<mac_addr, KismetClient::network_info *> prev_netmap;

// Core timer event that writes alerts out
int EmailEvent(Timetracker::timer_event *evt, void *parm,
					 GlobalRegistry *globalreg) {
	char cmdline[2048];
	FILE *mailer;

	map<mac_addr, KismetClient::network_info *> netmap = 
		globalreg->kisclient->FetchNetworks();

	if (netmap.size() <= prev_netmap.size())
		return 1;

	snprintf(cmdline, 2048, "%s -r \"%s\" %s",
			 mail_prog, from_addr, to_addr);

	if ((mailer = popen(cmdline, "w")) == NULL) {
		globalreg->messagebus->InjectMessage("Failed to open mailer program",
											 MSGFLAG_ERROR);
		return 1;
	}

	fprintf(mailer, "From: %s\n", from_addr);
	fprintf(mailer, "To: %s\n", to_addr);
	fprintf(mailer, "Subject: Kismet detected new network(s)\n");
	fprintf(mailer, "\n");

	fprintf(mailer, "Kismet has detected new network(s).\n\n");

	for (KismetClient::track_iter x = netmap.begin(); x != netmap.end(); ++x) {
		if (prev_netmap.find(x->first) != prev_netmap.end())
			continue;

		string typestr;
		switch (x->second->type) {
			case 0:
				typestr = "AP";
				break;
			case 1:
				typestr = "Ad-Hoc";
				break;
			case 2:
				typestr = "Probe";
				break;
			case 3:
				typestr = "Turbocell";
				break;
			case 4:
				typestr = "Data";
				break;
			default:
				typestr = "Unknown";
				break;
		}

		fprintf(mailer, "%.24s found new %s network %s \"%s\"%schannel %d\n",
				ctime((const time_t *) &(x->second->lasttime)),
				typestr.c_str(), x->second->bssid.Mac2String().c_str(),
				x->second->ssid == "" ? "<no ssid>" : x->second->ssid.c_str(),
				x->second->wep ? " Wep enabled " : " ",
				x->second->channel);
		
	}

	pclose(mailer);

	snprintf(cmdline, 2048, "Email triggered for %d new networks",
			 netmap.size() - prev_netmap.size());
	globalreg->messagebus->InjectMessage(cmdline, MSGFLAG_INFO);

	prev_netmap = netmap;
	
	return 1;
}

void Usage(char *argv) {
	fprintf(stderr, "%s [ -f configfile ]\n", argv);

	exit(1);
}

void CatchShutdown(int reason) {
	fprintf(stderr, "FATAL - Shutting down.\n");

	exit(1);
}

int main(int argc, char *argv[]) {
	char status[1024];

	static struct option long_options[] = {
		{ "config-file", required_argument, 0, 'f' },
	};

	char *configfile = NULL;

	int option_index;

	while (1) {
		int r = getopt_long(argc, argv, "f:o:",
							long_options, &option_index);

		if (r < 0)
			break;

		switch (r) {
			case 'f':
				configfile = optarg;
				break;

			default:
				Usage(argv[0]);
				break;
		}
	}
	
	// Build internal structures
	globalreg = new GlobalRegistry;
	
	// Create the messagebus and clients
	globalreg->messagebus = new MessageBus;
	StdoutMessageClient *stdmsgcli = new StdoutMessageClient(globalreg);
	globalreg->messagebus->RegisterClient(stdmsgcli, MSGFLAG_ALL);

	// Allocate a timetracker
	globalreg->timetracker = new Timetracker(globalreg);

	// Config file has to be opened before kisclient is created or it 
	// will be very sad.
	int freeconf = 0;
	if (configfile == NULL) {
		configfile = (char *) malloc(1024 * sizeof(char));
		snprintf(configfile, 1024, "%s/%s", SYSCONF_LOC, config_base);
		freeconf = 1;
	}

	ConfigFile *conf = new ConfigFile;

	// Parse the config file
	if (conf->ParseConfig(configfile) < 0) {
		exit(1);
	}

	// Assign it to the global registry
	globalreg->kismetui_config = conf;

	// Create the network client
	KismetClient *kis_cli = new KismetClient(globalreg);
	globalreg->kisclient = kis_cli;

	// Grab the network check rate
	int email_interval;
	if (conf->FetchOpt("email_interval") != "") {
		if (sscanf(conf->FetchOpt("email_interval").c_str(), "%d", 
				   &email_interval) != 1) {
			globalreg->messagebus->InjectMessage("Invalid email_interval in config "
												 "file.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;
			CatchShutdown(-1);
		}
	} else {
		globalreg->messagebus->InjectMessage("Missing email_rate in config",
											 MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		CatchShutdown(-1);
	}

	// Grab the mail program and addresses
	if (conf->FetchOpt("mailprog") == "") {
		globalreg->messagebus->InjectMessage("Missing mailprog in config file",
											 MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		CatchShutdown(-1);
	} else {
		mail_prog = strdup(conf->FetchOpt("mailprog").c_str());
	}

	if (conf->FetchOpt("fromaddress") == "") {
		globalreg->messagebus->InjectMessage("Missing fromaddress in config file",
											 MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		CatchShutdown(-1);
	} else {
		from_addr = strdup(conf->FetchOpt("fromaddress").c_str());
	}

	if (conf->FetchOpt("toaddress") == "") {
		globalreg->messagebus->InjectMessage("Missing toaddress in config file",
											 MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		CatchShutdown(-1);
	} else {
		to_addr = strdup(conf->FetchOpt("toaddress").c_str());
	}

	// Register a timer event for our alert printer
	globalreg->timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * email_interval,
										  NULL, 1, &EmailEvent, NULL);

	// Catch signals
	signal(SIGINT, CatchShutdown);
	signal(SIGTERM, CatchShutdown);
	signal(SIGHUP, CatchShutdown);

	while (1) {
		int max_fd = 0;
		fd_set rset, wset;
		struct timeval tm;

		FD_ZERO(&rset);
		FD_ZERO(&wset);

		max_fd = globalreg->kisclient->MergeSet(rset, wset, max_fd, &rset, &wset);

		tm.tv_sec = 0;
		tm.tv_usec = 100000;

		if (select(max_fd + 1, &rset, &wset, NULL, &tm) < 0) {
			if (errno != EINTR) {
				snprintf(status, STATUS_MAX, 
						 "select() error (%s)", strerror(errno));
				globalreg->messagebus->InjectMessage(status, MSGFLAG_FATAL);
				CatchShutdown(-1);
			}

		}

		// Only bail on fully fatal errors
		if (globalreg->kisclient->Poll(rset, wset) < 0 &&
			globalreg->fatal_condition)
			CatchShutdown(-1);

		globalreg->timetracker->Tick();

	}
	
}

